@@ -10,7 +10,6 @@
  * cva, 2002-10-26
  */
 
-#include "kpartx.h"
 #include <stdio.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -21,14 +20,16 @@
 #include <sys/types.h>
 #include <ctype.h>
 #include <libdevmapper.h>
+#include <linux/kdev_t.h>
 
+#include "devmapper.h"
 #include "crc32.h"
-
-/* loop devices */
 #include "lopart.h"
+#include "kpartx.h"
 
 #define SIZE(a) (sizeof(a)/sizeof((a)[0]))
 
+#define READ_SIZE	1024
 #define MAXTYPES	64
 #define MAXSLICES	256
 #define DM_TARGET	"linear"
@@ -85,78 +86,6 @@ usage(void) {
 	return 1;
 }
 
-static int
-dm_simplecmd (int task, const char *name) {
-	int r = 0;
-	struct dm_task *dmt;
-
-	if (!(dmt = dm_task_create (task)))
-		return 0;
-
-	if (!dm_task_set_name (dmt, name))
-		goto out;
-
-	r = dm_task_run (dmt);
-
-	out:
-	dm_task_destroy (dmt);
-	return r;
-}
-
-static int
-dm_addmap (int task, const char *name, const char *params, unsigned long size) {
-        struct dm_task *dmt;
-
-        if (!(dmt = dm_task_create (task)))
-                return 0;
-
-        if (!dm_task_set_name (dmt, name))
-                goto addout;
-
-        if (!dm_task_add_target (dmt, 0, size, DM_TARGET, params))
-                goto addout;
-
-        if (!dm_task_run (dmt))
-                goto addout;
-
-        addout:
-        dm_task_destroy (dmt);
-        return 1;
-}
-
-static int
-map_present (char * str)
-{
-	int r = 0;
-	struct dm_task *dmt;
-	struct dm_names *names;
-	unsigned next = 0;
-
-	if (!(dmt = dm_task_create (DM_DEVICE_LIST)))
-		return 0;
-
-        if (!dm_task_run (dmt))
-		goto out;
-
-	if (!(names = dm_task_get_names (dmt)))
-		goto out;
-
-	if (!names->dev)
-		goto out;
-
-	do {
-		if (0 == strcmp (names->name, str))
-			r = 1;
-
-		next = names->next;
-		names = (void *) names + next;
-	} while (next);
-
-	out:
-	dm_task_destroy (dmt);
-	return r;
-}
-
 static void
 set_delimiter (char * device, char * delimiter)
 {
@@ -194,20 +123,66 @@ find_devname_offset (char * device)
 
 	return (int)(q - device) + 1;
 }
-			
+
+static char *
+get_hotplug_device(void)
+{
+	unsigned int major, minor, off, len;
+	const char *mapname;
+	char *devname = NULL;
+	char *device = NULL;
+	char *var = NULL;
+	struct stat buf;
+
+	var = getenv("ACTION");
+
+	if (!var || strcmp(var, "add"))
+		return NULL;
+
+	/* Get dm mapname for hotpluged device. */
+	if (!(devname = getenv("DEVNAME")))
+		return NULL;
+
+	if (stat(devname, &buf))
+		return NULL;
+
+	major = (unsigned int)MAJOR(buf.st_rdev);
+	minor = (unsigned int)MINOR(buf.st_rdev);
+
+	if (!(mapname = dm_mapname(major, minor))) /* Not dm device. */
+		return NULL;
+
+	off = find_devname_offset(devname);
+	len = strlen(mapname);
+
+	/* Dirname + mapname + \0 */
+	if (!(device = (char *)malloc(sizeof(char) * (off + len + 1))))
+		return NULL;
+
+	/* Create new device name. */
+	snprintf(device, off + 1, "%s", devname);
+	snprintf(device + off, len + 1, "%s", mapname);
+
+	if (strlen(device) != (off + len))
+		return NULL;
+
+	return device;
+}
+
 int
 main(int argc, char **argv){
         int fd, i, j, k, n, op, off, arg;
 	struct slice all;
 	struct pt *ptp;
 	enum action what = LIST;
-	char *p, *type, *diskdevice, *device;
+	char *p, *type, *diskdevice, *device, *progname;
 	int lower, upper;
 	int verbose = 0;
 	char partname[PARTNAME_SIZE], params[PARTNAME_SIZE + 16];
 	char * loopdev = NULL;
 	char * delim = NULL;
 	int loopro = 0;
+	int hotplug = 0;
 	struct stat buf;
 
 	initpts();
@@ -218,7 +193,24 @@ main(int argc, char **argv){
 	memset(&all, 0, sizeof(all));
 	memset(&partname, 0, sizeof(partname));
 	
-	if (argc < 2) {
+	/* Check whether hotplug mode. */
+	progname = strrchr(argv[0], '/');
+
+	if (!progname)
+		progname = argv[0];
+	else
+		progname++;
+
+	if (!strcmp(progname, "kpartx.dev")) { /* Hotplug mode */
+		hotplug = 1;
+
+		/* Setup for original kpartx variables */
+		if (!(device = get_hotplug_device()))
+			exit(1);
+
+		diskdevice = device;
+		what = ADD;
+	} else if (argc < 2) {
 		usage();
 		exit(1);
 	}
@@ -258,7 +250,14 @@ main(int argc, char **argv){
 			exit(1);
 	}
 
-	if (optind == argc-2) {
+	if (dm_prereq(DM_TARGET, 0, 0, 0) && (what == ADD || what == DELETE)) {
+		fprintf(stderr, "device mapper prerequisites not met\n"); 
+		exit(1);
+	}
+
+	if (hotplug) {
+		/* already got [disk]device */
+	} else if (optind == argc-2) {
 		device = argv[optind];
 		diskdevice = argv[optind+1];
 	} else if (optind == argc-1) {
@@ -268,8 +267,8 @@ main(int argc, char **argv){
 		exit(1);
 	}
 
-	if (stat (device, &buf)) {
-		printf("failed to stat() device\n");
+	if (stat(device, &buf)) {
+		printf("failed to stat() %s\n", device);
 		exit (1);
 	}
 
@@ -367,7 +366,7 @@ main(int argc, char **argv){
 				}
 				strip_slash(partname);
 
-				if (!slices[j].size || !map_present(partname))
+				if (!slices[j].size || !dm_map_present(partname))
 					continue;
 
 				if (!dm_simplecmd(DM_DEVICE_REMOVE, partname))
@@ -406,10 +405,11 @@ main(int argc, char **argv){
 					exit(1);
 				}
 
-				op = (map_present(partname) ?
+				op = (dm_map_present(partname) ?
 					DM_DEVICE_RELOAD : DM_DEVICE_CREATE);
 
-				dm_addmap(op, partname, params, slices[j].size);
+				dm_addmap(op, partname, DM_TARGET, params,
+					  slices[j].size);
 
 				if (op == DM_DEVICE_RELOAD)
 					dm_simplecmd(DM_DEVICE_RESUME,
@@ -503,9 +503,9 @@ getblock (int fd, unsigned int secnr) {
 	bp->secnr = secnr;
 	bp->next = blockhead;
 	blockhead = bp;
-	bp->block = (char *) xmalloc(1024);
+	bp->block = (char *) xmalloc(READ_SIZE);
 	
-	if (read(fd, bp->block, 1024) != 1024) {
+	if (read(fd, bp->block, READ_SIZE) != READ_SIZE) {
 		fprintf(stderr, "read error, sector %d\n", secnr);
 		bp->block = NULL;
 	}
